# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
import mock
import os
import stat
from textwrap import dedent
from unittest import TestCase

from mojo.phase import (
    CharmAuditPhase,
    CharmRepoPhase,
    CollectPhase,
    CollectPhaseException,
    DeployerPhase,
    )
from mojo.tests import utils
from mojo.project import Project


class DeployerPhaseTestCase(TestCase, utils.MojoTestCaseMixin):

    def get_deployer_command_for_run(self, phase, workspace, stage='staging'):
        """Return the deployer command executed when the phase runs."""
        # XXX: It'd be nicer to stub mojo.utils.subprocess.Popen
        # but bicommand makes that quite complicated.
        with mock.patch('mojo.phase.bicommand') as mock_bicommand:
            with mock.patch('mojo.phase.DeployerPhase._confirm_charm_repo'):
                with mock.patch('mojo.phase.DeployerPhase._get_target', return_value="testTarget"):
                    with mock.patch('mojo.juju.Status.check_and_wait'):
                        mock_bicommand.return_value = (0, 'fake-output')

                        project = Project('unused', 'trusty', '/srv/mojo')
                        phase.run(project, workspace, stage)

                        return mock_bicommand.call_args[0][0]

    def test_default_config_rendered_for_deployer(self):
        phase = DeployerPhase()
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['staging/deploy', 'deploy'],
            stage='staging')

        command = self.get_deployer_command_for_run(phase, workspace)

        # Verify that the configs were both rendered.
        self.assertTrue(os.path.exists(
            '{}/deploy'.format(workspace.repo_dir)))
        self.assertTrue(os.path.exists(
            '{}/staging/deploy'.format(workspace.repo_dir)))
        # Verify that the configs were included in the deployer command.
        expected_configs = (
            '-c {configs_dir}/deploy '
            '-c {configs_dir}/staging/deploy'.format(
                configs_dir=workspace.repo_dir))
        self.assertIn(expected_configs, command)

    def test_multiple_configs_rendered_for_deployer(self):
        phase = DeployerPhase(config='services.cfg')
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['staging/services.cfg', 'services.cfg'],
            stage='staging')

        command = self.get_deployer_command_for_run(phase, workspace)

        # Verify that the configs were both rendered.
        self.assertTrue(os.path.exists(
            '{}/services.cfg'.format(workspace.repo_dir)))
        self.assertTrue(os.path.exists(
            '{}/staging/services.cfg'.format(workspace.repo_dir)))
        # Verify that the configs were included in the deployer command.
        expected_configs = (
            '-c {configs_dir}/services.cfg '
            '-c {configs_dir}/staging/services.cfg'.format(
                configs_dir=workspace.repo_dir))
        self.assertIn(expected_configs, command)

    def test_comma_separated_configs_rendered_for_deployer(self):
        phase = DeployerPhase(config='services.cfg,placement.cfg')
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['staging/services.cfg', 'staging/placement.cfg', 'services.cfg'],
            stage='staging')

        command = self.get_deployer_command_for_run(phase, workspace)

        # Verify that the configs were rendered.
        self.assertTrue(os.path.exists(
            '{}/services.cfg'.format(workspace.repo_dir)))
        self.assertTrue(os.path.exists(
            '{}/staging/services.cfg'.format(workspace.repo_dir)))
        self.assertTrue(os.path.exists(
            '{}/staging/placement.cfg'.format(workspace.repo_dir)))
        # Verify that the configs were included in the deployer command.
        expected_configs = (
            '-c {configs_dir}/services.cfg '
            '-c {configs_dir}/staging/services.cfg '
            '-c {configs_dir}/staging/placement.cfg'.format(
                configs_dir=workspace.repo_dir))
        self.assertIn(expected_configs, command)

    def test_local_config_rendered_with_same_perms(self):
        phase = DeployerPhase(config='services', local="local.yaml")
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['services'],
            stage='staging')
        utils.touch(os.path.join(workspace.local_dir, "local.yaml"))

        command = self.get_deployer_command_for_run(phase, workspace)

        expected_configs = (
            '-c {configs_dir}/services '
            '-c {configs_dir}/local.yaml-local.cfg'.format(
                configs_dir=workspace.repo_dir))
        self.assertIn(expected_configs, command)
        mode = os.stat(os.path.join(workspace.repo_dir, 'local.yaml-local.cfg')).st_mode
        group_other_bits = (stat.S_IROTH | stat.S_IWOTH | stat.S_IXOTH |
                            stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP)

        self.assertFalse(mode & group_other_bits,
                         "The rendered local config is not accessible to "
                         "group or others.")

    def test_multiple_local_config_rendered_with_same_perms(self):
        local = "secrets,local.yaml"
        phase = DeployerPhase(config='services', local=local)
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['services'],
            stage='staging')
        utils.touch(os.path.join(workspace.local_dir, "secrets"))
        utils.touch(os.path.join(workspace.local_dir, "local.yaml"))

        command = self.get_deployer_command_for_run(phase, workspace)

        expected_configs = (
            '-c {configs_dir}/services '
            '-c {configs_dir}/secrets-local.cfg '
            '-c {configs_dir}/local.yaml-local.cfg'.format(
                configs_dir=workspace.repo_dir))
        self.assertIn(expected_configs, command)

        group_other_bits = (stat.S_IROTH | stat.S_IWOTH | stat.S_IXOTH |
                            stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP)

        for config in local.split(','):
            path = os.path.join(workspace.repo_dir,
                                '{}-local.cfg'.format(config))
            mode = os.stat(path).st_mode

            self.assertFalse(mode & group_other_bits,
                             "The rendered local config for {} is not "
                             "accessible to group or others.".format(config))

    @mock.patch('mojo.phase.SecretsPhase')
    def test_secrets_automatically_run(self, secrets):
        phase = DeployerPhase(config='services.cfg', target='test')
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['staging/services.cfg', 'services.cfg'],
            stage='staging')

        self.get_deployer_command_for_run(phase, workspace)

        secrets.assert_called_with(auto_secrets=True)
        called_methods = [c[0] for c in secrets.mock_calls]
        assert ('().run' in called_methods)


class CollectPhaseTestCase(TestCase, utils.MojoTestCaseMixin):

    def setUp(self):
        self.empty_project = Project('unused', 'trusty', '/tmp')

    @mock.patch('codetree.config.Config')
    def test_codetree_is_called(self, codetree_Config):
        phase = CollectPhase()
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['collect', 'services.cfg'], stage='staging')

        phase.run(self.empty_project, workspace, 'staging')

        called_methods = [c[0] for c in codetree_Config.mock_calls]
        assert ('().build' in called_methods)

    @mock.patch('codetree.config.Config')
    def test_codetree_is_called_with_workers(self, codetree_Config):
        phase = CollectPhase()
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['collect', 'services.cfg'], stage='staging')

        # With no workers specified to start with
        phase.run(self.empty_project, workspace, 'staging')
        called_without_workers = 'call().build()' in \
            [str(c) for c in codetree_Config.mock_calls]
        assert (called_without_workers)

        # And now with workers specified
        phase.options.update({'workers': 2})
        phase.run(self.empty_project, workspace, 'staging')

        called_with_workers = 'call().build(workers=2)' in \
            [str(c) for c in codetree_Config.mock_calls]
        assert (called_with_workers)

    @mock.patch('codetree.config.Config')
    def test_codetree_is_called_with_workers_unsupported(self, codetree_Config):
        phase = CollectPhase()
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['collect', 'services.cfg'], stage='staging')

        phase.options.update({'workers': 2})
        errmsg = "build() got an unexpected keyword argument 'workers'"
        codetree_Config.return_value.build.side_effect = TypeError(errmsg)

        with self.assertRaises(CollectPhaseException):
            phase.run(self.empty_project, workspace, 'staging')

        called_with_workers = 'call().build(workers=2)' in \
            [str(c) for c in codetree_Config.mock_calls]
        assert (called_with_workers)

    @mock.patch('codetree.config.Config')
    def test_codetree_failure_exception(self, _codetree_Config):
        _codetree_Config.return_value.build.return_value = False
        phase = CollectPhase()
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['collect', 'services.cfg'], stage='staging')

        with self.assertRaises(CollectPhaseException):
            phase.run(self.empty_project, workspace, 'staging')

    @mock.patch('mojo.phase.CharmRepoPhase')
    @mock.patch('codetree.config.Config')
    def test_repo_automatically_runs(self, codetree_Config, charm_repo):
        phase = CollectPhase()
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['collect', 'services.cfg'], stage='staging')

        phase.run(self.empty_project, workspace, 'staging')

        called_methods = [c[0] for c in charm_repo.mock_calls]
        assert ('().run' in called_methods)


class CharmAuditPhaseTestCase(TestCase, utils.MojoTestCaseMixin):

    def test_parse_charm_audit_output(self):
        # In this example, landscape-client and ntp don't have a codetree-collect-info.yaml
        # file, but telegraf does.
        audit_output = dedent("""\
            landscape-client-1:
              application: landscape-client
            ntp-0:
              application: ntp
            telegraf-12:
              application: telegraf
              collect_date: '2017-11-01 11:34:44.479841'
              collect_url: git://git.launchpad.net/telegraf-charm;revno=ad93f7333da35a
              hashes:
                COPYING.txt: 8ceb4b9ee5adedde47b31e975c1d90c73ad27b6b165a1dcd80c7c545eb65b903
            """)
        expected_output = {
            "landscape-client": {"Unknown": ["landscape-client-1"]},
            "ntp": {"Unknown": ["ntp-0"]},
            "telegraf": {"git://git.launchpad.net/telegraf-charm;revno=ad93f7333da35a": ["telegraf-12"]},
            }
        phase = CharmAuditPhase()
        self.assertEqual(phase.parse_charm_audit_output(audit_output), expected_output)

    def test_charm_audit_summary(self):
        # In this example, landscape-client and ntp don't have a codetree-collect-info.yaml
        # file, but telegraf does.
        parsed_audit_output = {
            "landscape-client": {"Unknown": ["landscape-client-1"]},
            "ntp": {"Unknown": ["ntp-0"]},
            "telegraf": {"git://git.launchpad.net/telegraf-charm;revno=ad93f7333da35a": ["telegraf-12"]},
            }
        expected_output = dedent("""\
            Application: landscape-client
              Unable to determine charm version for landscape-client-1
            Application: ntp
              Unable to determine charm version for ntp-0
            Application: telegraf
              The charm git://git.launchpad.net/telegraf-charm;revno=ad93f7333da35a is deployed on telegraf-12
            """).strip()
        phase = CharmAuditPhase()
        self.assertEqual("\n".join(phase.charm_audit_summary(parsed_audit_output)), expected_output)


class CharmRepoPhaseTestCase(TestCase, utils.MojoTestCaseMixin):

    def setUp(self):
        self.empty_project = Project('unused', 'trusty', '/tmp')

    @mock.patch('mojo.charm_repo.CharmRepo.build_from_config')
    @mock.patch('mojo.charm_repo.CharmRepo.build_all')
    def test_repo_default(self, build_all, build_from_config):
        phase = CharmRepoPhase()
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['repo', 'services.cfg'], stage='staging')

        phase.run(self.empty_project, workspace, 'staging')

        build_from_config.assert_called_with(
            os.path.join(workspace.spec_dir, 'repo'),
            os.path.join(
                os.path.dirname(workspace.spec_dir), 'build'),
            'trusty')

    @mock.patch('mojo.charm_repo.CharmRepo.build_from_config')
    @mock.patch('mojo.charm_repo.CharmRepo.build_all')
    def test_repo_with_config(self, build_all, build_from_config):
        phase = CharmRepoPhase(config='repo-update')
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['repo-update', 'services.cfg'], stage='staging')

        phase.run(self.empty_project, workspace, 'staging')

        build_from_config.assert_called_with(
            os.path.join(workspace.spec_dir, 'repo-update'),
            os.path.join(
                os.path.dirname(workspace.spec_dir), 'build'),
            'trusty')

    @mock.patch('mojo.charm_repo.CharmRepo.build_from_config')
    @mock.patch('mojo.charm_repo.CharmRepo.build_all')
    def test_auto_repo(self, build_all, build_from_config):
        phase = CharmRepoPhase(auto_repo=True)
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['repo', 'services.cfg'], stage='staging')

        phase.run(self.empty_project, workspace, 'staging')
        build_all.assert_called_with(
            os.path.join(os.path.dirname(workspace.spec_dir), 'build'),
            'trusty')
